from os import PathLike
import json

import numpy
import h5py

from ..tasks.integrate import MultiConfigIntegrateSinglePattern
from ..tasks.nexus import SaveNexusMultiPattern1D

from .xrpd_theory import Setup
from .xrpd_theory import Measurement
from .xrpd_theory import RadialPattern


def test_multiintegrate1d(
    imageSetup1SampleA: Measurement, setup1: Setup, xSampleA: RadialPattern
):
    configs = [
        {"nbpt_rad": 50, "unit": "2th_rad"},
        {"nbpt_rad": 50, "unit": "q_A^-1"},
        {"nbpt_rad": 100, "unit": "q_A^-1"},
    ]
    integration_options = {**setup1.integration_options, **xSampleA.integration_options}
    inputs = {
        "image": imageSetup1SampleA.image,
        "detector": setup1.detector,
        "detector_config": setup1.detector_config,
        "geometry": setup1.geometry,
        "energy": setup1.energy,
        "integration_options": integration_options,
        "configs": configs,
    }

    task = MultiConfigIntegrateSinglePattern(inputs=inputs)
    task.execute()
    output_values = task.get_output_values()

    assert len(output_values["radial_units"]) == 3
    for unit, config in zip(output_values["radial_units"], configs):
        assert unit == config["unit"]

    assert len(output_values["radial"]) == 3
    for x, config in zip(output_values["radial"], configs):
        assert len(x) == config["nbpt_rad"]

    assert len(output_values["intensity"]) == 3
    for y, config in zip(output_values["intensity"], configs):
        assert len(y) == config["nbpt_rad"]

    assert len(output_values["info"]) == 3
    for info, config in zip(output_values["info"], configs):
        assert info["nbpt_rad"] == config["nbpt_rad"]
        assert info["unit"] == config["unit"]


def test_multiintegrate2d(
    imageSetup1SampleA: Measurement, setup1: Setup, xSampleA: RadialPattern
):
    configs = [
        {"nbpt_rad": 100, "unit": "q_A^-1"},
        {"nbpt_rad": 100, "unit": "q_A^-1", "nbpt_azim": 360},
    ]
    integration_options = {**setup1.integration_options, **xSampleA.integration_options}
    inputs = {
        "image": imageSetup1SampleA.image,
        "detector": setup1.detector,
        "detector_config": setup1.detector_config,
        "geometry": setup1.geometry,
        "energy": setup1.energy,
        "integration_options": integration_options,
        "configs": configs,
    }

    task = MultiConfigIntegrateSinglePattern(inputs=inputs)
    task.execute()
    output_values = task.get_output_values()

    assert len(output_values["radial_units"]) == 2
    assert output_values["radial_units"][0] == "q_A^-1"
    assert output_values["radial_units"][1] == "q_A^-1"

    assert len(output_values["radial"]) == 2
    assert output_values["radial"][0].shape == (100,)
    assert output_values["radial"][1].shape == (100,)

    assert len(output_values["azimuthal"]) == 2
    assert output_values["azimuthal"][0] is None
    assert output_values["azimuthal"][1].shape == (360,)

    assert len(output_values["intensity"]) == 2
    assert output_values["intensity"][0].shape == (100,)
    assert output_values["intensity"][1].shape == (360, 100)

    assert len(output_values["info"]) == 2
    assert output_values["info"][0]["nbpt_rad"] == 100
    assert "nbpt_azim" not in output_values["info"][0]
    assert output_values["info"][1]["nbpt_rad"] == 100
    assert output_values["info"][1]["nbpt_azim"] == 360


def test_multisave(tmpdir: PathLike, setup1: Setup):
    npt_list = (50, 100, 200)
    inputs = {
        "url": str(tmpdir / "result.h5"),
        "x_list": [numpy.arange(npt) for npt in npt_list],
        "y_list": [numpy.random.random(npt) for npt in npt_list],
        "xunits": ["2th_deg", "q_A^-1", "q_A^-1"],
        "header_list": [
            {
                "energy": 10.2,
                "detector": setup1.detector,
                "detector_config": setup1.detector_config,
                "geometry": setup1.geometry,
                "npt_radial": npt,
            }
            for npt in npt_list
        ],
    }

    task = SaveNexusMultiPattern1D(inputs=inputs)
    task.execute()

    with h5py.File(str(tmpdir / "result.h5")) as root:
        assert set(root["results"].keys()) == {
            "measurement",
            "integrate_0",
            "integrate_1",
            "integrate_2",
            # "instrument",
            # "dummy",
        }

        for i, (npt, unit) in enumerate(zip(npt_list, inputs["xunits"])):
            nxprocess = root[f"results/integrate_{i}"]
            numpy.testing.assert_array_equal(
                nxprocess[f"integrated/{unit.split('_')[0]}"],
                inputs["x_list"][i],
            )
            numpy.testing.assert_array_equal(
                nxprocess["integrated/intensity"], inputs["y_list"][i]
            )
            numpy.testing.assert_array_equal(
                root[f"results/measurement/integrated_{i}"], inputs["y_list"][i]
            )

            config = json.loads(nxprocess["configuration"]["data"][()])
            assert set(config) == {
                "ewoks_version",
                "energy",
                "detector",
                "detector_config",
                "geometry",
                "npt_radial",
            }
            numpy.testing.assert_array_equal(config["npt_radial"], npt)



import re
import logging

from ..tasks.log import zip_with_progress


def test_zip_with_progress_one_iterator(caplog):
    with caplog.at_level(logging.INFO):
        values = list()
        for value, *_ in zip_with_progress([], message=_MESSAGE_TEMPLATE):
            values.append(value)

    assert not values
    _assert_finished_log(caplog, 0)

    with caplog.at_level(logging.INFO):
        values = list()
        for value, *_ in zip_with_progress([1], message=_MESSAGE_TEMPLATE):
            values.append(value)

    assert values == [1]
    _assert_finished_log(caplog, 1)

    with caplog.at_level(logging.INFO):
        values = list()
        for value, *_ in zip_with_progress([1, 2], message=_MESSAGE_TEMPLATE):
            values.append(value)

    assert values == [1, 2]
    _assert_finished_log(caplog, 2)


def test_zip_with_progress_two_iterators(caplog):
    with caplog.at_level(logging.INFO):
        values = list()
        for value, *_ in zip_with_progress([], [1, 2], message=_MESSAGE_TEMPLATE):
            values.append(value)

    assert not values
    _assert_finished_log(caplog, 0)

    with caplog.at_level(logging.INFO):
        values = list()
        for nr, letter, *_ in zip_with_progress(
            [1, 2], ["a"], message=_MESSAGE_TEMPLATE
        ):
            values.append((nr, letter))

    assert values == [(1, "a")]
    _assert_finished_log(caplog, 1)


_MESSAGE_TEMPLATE = "Integrated %s images of %s"


def _assert_finished_log(caplog, nimages):
    pattern = rf"Integrated {nimages} images of \? \(FINISHED: iteration=\d{{2}}:\d{{2}}:\d{{2}} \(\d{{1,3}}%\), TOTAL=\d{{2}}:\d{{2}}:\d{{2}} \(\d{{1,3}}%\)\)"
    string = str(caplog.text)
    err_msg = f"'{string}' does not match '{pattern}'"
    assert re.search(pattern, string), err_msg



from typing import Any, Optional
from dataclasses import dataclass, field

import numpy
import pyFAI.units
import pyFAI.calibrant
from numpy.typing import ArrayLike

from .. import pyfai_api


@dataclass(frozen=True)
class RadialPattern:
    x: ArrayLike = field(repr=False)
    low_limit: float
    high_limit: float
    units: str

    @property
    def integration_options(self) -> dict:
        return {
            "radial_range_min": self.low_limit,
            "radial_range_max": self.high_limit,
            "nbpt_rad": self.x.size,
            "unit": self.units,
        }


@dataclass(frozen=True)
class AzimuthalPattern:
    x: ArrayLike = field(repr=False)
    low_limit: float
    high_limit: float
    units: str = "chi_deg"

    @property
    def integration_options(self) -> dict:
        return {
            "azimuth_range_min": self.low_limit,
            "azimuth_range_max": self.high_limit,
            "nbpt_azim": self.x.size,
        }


@dataclass(frozen=True)
class IntensityPattern:
    y: ArrayLike = field(repr=False)
    monitor: float
    theory: Any


@dataclass(frozen=True)
class Measurement:
    image: ArrayLike = field(repr=False)
    monitor: float


@dataclass(frozen=True)
class Calibration:
    image: ArrayLike = field(repr=False)
    calibrant: str


@dataclass(frozen=True)
class Setup:
    version: Optional[int]
    detector: str
    energy: float  # keV
    geometry: dict

    @property
    def wavelength(self) -> float:  # Angstrom
        return pyFAI.units.hc / (self.energy * 1e10)

    @property
    def detector_config(self) -> dict:
        return {}

    def load_detector_config(self) -> dict:
        detector_object = pyFAI.detectors.detector_factory(self.detector)
        return detector_object.get_config()

    @property
    def integration_options(self) -> dict:
        if not self.version:
            return {}
        return {"version": self.version}


def calibration(
    name: str, ai: pyfai_api.AzimuthalIntegrator, setup: Setup
) -> Calibration:
    calibrant = pyFAI.calibrant.get_calibrant(name)
    calibrant.set_wavelength(setup.wavelength)
    # W = FWHM^2 in rad
    FWHM = 0.2  # deg
    W = numpy.radians(FWHM) ** 2
    image = calibrant.fake_calibration_image(ai, Imax=100, W=W)
    return Calibration(image=image, calibrant=name)


def measurement(
    ai: pyfai_api.AzimuthalIntegrator,
    xpattern,
    ypattern,
    mult=2,
):
    x = xpattern.x
    y = mult * ypattern.y
    monitor = mult * ypattern.monitor
    image = ai.calcfrom1d(x, y, dim1_unit=xpattern.units, mask=ai.mask)
    return Measurement(image=image, monitor=monitor)


def measurement2d(
    ai: pyfai_api.AzimuthalIntegrator,
    radial: RadialPattern,
    azimuthal: AzimuthalPattern,
    intensity: IntensityPattern,
):
    image = ai.calcfrom2d(
        intensity.y,
        radial.x,
        azimuthal.x,
        dim1_unit=radial.units,
        dim2_unit=azimuthal.units,
        mask=ai.mask,
    )
    return Measurement(image=image, monitor=intensity.monitor)



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator, Union
import h5py
import numpy
from silx.io.nxdata import NXdata


@dataclass
class IntegratedPattern:
    """Store one pyFAI integrated pattern"""

    point: Union[float, int, None]
    radial: numpy.ndarray
    radial_units: str
    intensity: numpy.ndarray
    intensity_errors: Union[numpy.ndarray, None]


def read_nexus_integrated_patterns(group: h5py.Group) -> Generator[IntegratedPattern]:
    """Read integrated patterns from a HDF5 NXdata group.

    It reads from both single (1D signal) or multi (2D signal) NXdata.
    """
    nxdata = NXdata(group)
    if not nxdata.is_valid:
        raise RuntimeError(
            f"Cannot parse NXdata group: {group.file.filename}::{group.name}"
        )
    if not (nxdata.signal_is_1d or nxdata.signal_is_2d):
        raise RuntimeError(
            f"Signal is not a 1D or 2D dataset: {group.file.filename}::{group.name}"
        )

    if nxdata.signal_is_1d:
        points = [None]
    else:  # 2d
        if nxdata.axes[0] is None:
            points = [None] * nxdata.signal.shape[0]
        else:
            points = nxdata.axes[0][()]

    if nxdata.axes[-1] is None:
        radial = numpy.arange(nxdata.signal.shape[1])
        radial_units = ""
    else:
        axis_dataset = nxdata.axes[-1]
        radial = axis_dataset[()]
        radial_units = axis_dataset.attrs.get("units", "")

    intensities = numpy.atleast_2d(nxdata.signal)

    if nxdata.errors is None:
        errors = [None] * intensities.shape[0]
    else:
        errors = numpy.atleast_2d(nxdata.errors)

    if (len(points), len(radial)) != intensities.shape:
        raise RuntimeError("Shape mismatch between axes and signal")

    for point, intensity, intensity_errors in zip(points, intensities, errors):
        yield IntegratedPattern(
            point, radial, radial_units, intensity, intensity_errors
        )



from __future__ import annotations

import logging
import os
import zipfile
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Mapping
from typing import Sequence
from typing import Tuple

import numpy
from ewokscore import Task

from .data_access import TaskWithDataAccess
from .utils.ascii_utils import ensure_parent_folder
from .utils.ascii_utils import save_pattern_as_ascii
from .utils.nexus_utils import IntegratedPattern
from .utils.nexus_utils import read_nexus_integrated_patterns

logger = logging.getLogger(__name__)


__all__ = [
    "SaveAsciiPattern1D",
    "SaveAsciiMultiPattern1D",
    "SaveNexusPatternsAsAscii",
]


class SaveAsciiPattern1D(
    Task,
    input_names=["filename", "x", "y", "xunits"],
    optional_input_names=["header", "yerror", "metadata"],
    output_names=["saved"],
):
    """Save single diffractogram in ASCII format"""

    def run(self):
        header = self.get_input_value("header", dict())
        metadata = self.get_input_value("metadata", dict())
        yerror = self.get_input_value("yerror", None)

        save_pattern_as_ascii(
            self.inputs.filename,
            self.inputs.x,
            self.inputs.y,
            self.inputs.xunits,
            yerror,
            header,
            metadata,
        )
        self.outputs.saved = True


class SaveAsciiMultiPattern1D(
    Task,
    input_names=["filenames", "x_list", "y_list", "xunits_list"],
    optional_input_names=["header_list", "yerror_list", "metadata_list"],
    output_names=["saved"],
):
    def run(self):
        filenames: Sequence[str] = self.inputs.filenames
        x_list: Sequence[numpy.ndarray] = self.inputs.x_list
        y_list: Sequence[numpy.ndarray] = self.inputs.y_list
        xunits_list: Sequence[str] = self.inputs.xunits_list
        header_list: Sequence[Mapping] = self.get_input_value(
            "header_list", len(filenames) * [dict()]
        )
        yerror_list: Sequence[numpy.ndarray] | Sequence[None] = self.get_input_value(
            "yerror_list", len(filenames) * [None]
        )
        metadata_list: Sequence[Dict[str, Any]] = self.get_input_value(
            "metadata_list", len(filenames) * [dict()]
        )

        for args in zip(
            filenames,
            x_list,
            y_list,
            xunits_list,
            yerror_list,
            header_list,
            metadata_list,
        ):
            save_pattern_as_ascii(*args)

        self.outputs.saved = True


class SaveNexusPatternsAsAscii(
    TaskWithDataAccess,
    input_names=["nxdata_url", "output_filename_template"],
    optional_input_names=["header", "enabled", "output_archive_filename", "overwrite"],
    output_names=["filenames"],
):
    """Convert azimuthal integration results from NeXus to ASCII files

    One ASCII file is created for each integration pattern in nxdata_url.
    ASCII files are named from the output_filename_template and the pattern index.

    If output_archive_filename is provided, all ASCII files are stored in a single ZIP file.
    In this case, output_filename_template is the template path of ASCII files inside the ZIP file.

    Required inputs:
    - nxdata_url (str): The url of the NXData group storing the azimuthal integration results
    - output_filename_template (str): A string template containing one '%d' field.
      It is used to generate the filename from the frame number.

    Optional inputs:
    - header (dict): Information to store in ASCII file header (default: {})
    - enabled (bool): True to enable saving as ASCII files, False to skip task (default: True)
    - output_archive_filename (str): Filename of the ZIP file containing all ASCII files.
      If this is None (default) or the empty string, ZIP compression is disabled.
    - overwrite (bool): True to allow overwriting existing ASCII/ZIP files (default: False)

    Outputs:
    - filenames (tuple[str]): The names of the created ASCII files or ZIP file
    """

    def run(self):
        if not self.get_input_value("enabled", True):
            logger.info(
                f"Task {self.__class__.__qualname__} is disabled: No file is saved"
            )
            self.outputs.filenames = tuple()
            return

        output_archive_filename = self.get_input_value("output_archive_filename", None)
        overwrite = self.get_input_value("overwrite", False)
        if output_archive_filename:
            ensure_parent_folder(output_archive_filename)

            mode = "w" if overwrite else "x"
            with zipfile.ZipFile(
                output_archive_filename, mode=mode, compression=zipfile.ZIP_DEFLATED
            ) as zipf:
                with self.open_h5item(self.inputs.nxdata_url) as group:
                    for filename, pattern, metadata in self._export_data(group):
                        with zipf.open(filename, mode="w") as file:
                            save_pattern_as_ascii(
                                file,
                                x=pattern.radial,
                                y=pattern.intensity,
                                xunits=pattern.radial_units,
                                yerror=pattern.intensity_errors,
                                header=self.get_input_value("header", {}),
                                metadata=metadata,
                            )
            self.outputs.filenames = (output_archive_filename,)
        else:
            filenames = []
            with self.open_h5item(self.inputs.nxdata_url) as group:
                for filename, pattern, metadata in self._export_data(group):
                    if not overwrite and os.path.exists(filename):
                        raise FileExistsError(f"File exists: {filename}")
                    save_pattern_as_ascii(
                        filename,
                        x=pattern.radial,
                        y=pattern.intensity,
                        xunits=pattern.radial_units,
                        yerror=pattern.intensity_errors,
                        header=self.get_input_value("header", {}),
                        metadata=metadata,
                    )
                    filenames.append(filename)
                self.outputs.filenames = tuple(filenames)

    def _export_data(self, group) -> Iterator[Tuple[str, IntegratedPattern, dict]]:
        for index, pattern in enumerate(read_nexus_integrated_patterns(group)):
            filename = self.inputs.output_filename_template % index

            if pattern.point is not None:
                metadata = {"point": pattern.point}
            else:
                metadata = {}

            yield filename, pattern, metadata



import os
import time
from contextlib import ExitStack
from typing import Optional, Iterator, Generator, Tuple, Any

import h5py
from ewokscore import Task
from blissdata.h5api import dynamic_hdf5

from .data_access import TaskWithDataAccess
from .utils.sum_utils import save_sum
from .utils.sum_utils import ImageSum
from .utils.sum_utils import generate_range
from .utils.data_utils import hdf5_url

SUM_TYPES = ["per_scan", "all_scans", "both"]


class SumBlissScanImages(
    TaskWithDataAccess,
    input_names=["filename", "scan", "detector_name", "output_filename"],
    optional_input_names=[
        "monitor_name",
        "subscan",
        "scan_memory_url",
        "output_process",
        "background_step",
    ],
    output_names=["output_uri", "monitor"],
):
    """Sum images of a single camera of a single Bliss scan"""

    def run(self):
        with ExitStack() as stack:
            self._run_in_stack(stack)

    def _run_in_stack(self, stack: ExitStack):
        filename: str = self.inputs.filename
        scan: int = self.inputs.scan
        subscan = self.get_input_value("subscan", 1)
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        output_nxentry_url = hdf5_url(output_filename, f"/{scan}.{subscan}")
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_nxprocess_name: str = self.get_input_value("output_process", "sum")
        background_step: bool = self.get_input_value("background_step", 1)

        if self.inputs.scan_memory_url:
            raise NotImplementedError("data from memory not supported yet")
        else:
            lima_iterator_context = self.iter_bliss_data(
                filename,
                scan,
                lima_names=[detector_name],
                subscan=subscan,
            )
            lima_iterator = stack.enter_context(lima_iterator_context)

            with dynamic_hdf5.File(
                filename, lima_names=[detector_name], **self.get_retry_options()
            ) as root:
                data_name = f"{scan}.{subscan}/instrument/{detector_name}/data"
                # getting the shape should block until the end of the scan
                dataset_shape = root[data_name].shape
                nb_points_in_scan = dataset_shape[0]

            if monitor_name:
                counter_iterator_context = self.iter_bliss_data(
                    filename,
                    scan,
                    counter_names=[monitor_name],
                    lima_names=[],
                    subscan=subscan,
                )
                counter_iterator = stack.enter_context(counter_iterator_context)
            else:
                counter_iterator = (
                    {monitor_name: None} for _ in range(nb_points_in_scan)
                )

        with self.open_h5item(output_nxentry_url, mode="a", create=True) as nxentry:
            nxprocess = nxentry.create_group(output_nxprocess_name)
            nxprocess.attrs["NX_class"] = "NXprocess"
            nxprocess.attrs["default"] = "results"
            nxdata = nxprocess.create_group("results")
            nxdata.attrs["NX_class"] = "NXdata"

            scan_sum = None
            summed_indices = []

            tstart = time.time()
            for scan_index, ctrdata in iterate_scan_with_skip(
                counter_iterator, background_step
            ):
                limadata = next(lima_iterator)

                image = limadata[detector_name]
                if scan_sum is None:
                    scan_sum = ImageSum(image.shape)

                scan_sum.add_to_sum(image, monitor=ctrdata[monitor_name])
                summed_indices.append(scan_index)

            if scan_sum is None:
                tend = time.time()
                raise RuntimeError(
                    f"No scan data yielded within {tend-tstart:03f} from {filename}::/{scan}.{subscan}"
                )

            image_dset, monitor_dset = save_sum(
                nxdata,
                name=f"Scan{scan}-Images{summed_indices[0]}-{summed_indices[-1]}",
                image_sum=scan_sum,
            )

            self.outputs.output_uri = f"{nxdata.file.filename}::{image_dset.name}"
            if monitor_dset is not None:
                self.outputs.monitor = scan_sum.summed_monitor
            else:
                self.outputs.monitor = None


def iterate_scan_with_skip(
    scan_iterator: Iterator[Any], step: int
) -> Generator[Tuple[int, Any], None, None]:
    """
    :param scan_iterator: iterate over all scan points
    :param step: -1 means no skipping, 0 means skipping the first, N>0 means skipping every N points starting from 0
    :yields: the non-skipped scan index and corresponding data
    """
    if step < 0:  # no skipping
        for scan_index, data in enumerate(scan_iterator):
            yield scan_index, data
    elif step == 0:  # skip the first
        for scan_index, data in enumerate(scan_iterator):
            if scan_index == 0:
                continue
            yield scan_index, data
    else:  # skip every N points, starting from zero
        block_size = step + 1
        for scan_index, data in enumerate(scan_iterator):
            if (scan_index % block_size) == 0:
                continue
            yield scan_index, data


class SumImages(
    Task,
    input_names=["filename", "detector_name", "output_filename"],
    optional_input_names=[
        "start_scan",
        "end_scan",
        "start_image",
        "end_image",
        "block_size",
        "monitor_name",
        "output_entry",
        "output_process",
        "sum_type",
    ],
    output_names=["output_uris", "monitor_uris"],
):
    """Sum images of a single camera from a Bliss scan file

    For each scan, images are added in blocks of `block_size` images (one block with all images by default).

    The result contains:
        * the block sums when sum_type=per_scan or sum_type=both
        * the sum of the block sums when sum_type=all_scans or sum_type=both
    """

    def run(self):
        filename: str = self.inputs.filename
        detector_name: str = self.inputs.detector_name
        output_filename: str = self.inputs.output_filename
        start_scan: int = self.get_input_value("start_scan", 1)
        end_scan: Optional[int] = self.get_input_value("end_scan", None)
        start_image: int = self.get_input_value("start_image", 0)
        end_image: Optional[int] = self.get_input_value("end_image", None)
        block_size: Optional[int] = self.get_input_value("block_size", None)
        monitor_name: Optional[str] = self.get_input_value("monitor_name", None)
        output_entry: str = self.get_input_value("output_entry", "processing")
        output_process: str = self.get_input_value("output_process", "sum")
        sum_type: str = self.get_input_value("sum_type", "per_scan")

        if sum_type not in SUM_TYPES:
            raise TypeError(
                f"sum_type must be one of the following values: {SUM_TYPES}. Got {sum_type} instead."
            )
        save_scan_sums = sum_type == "per_scan" or sum_type == "both"
        save_full_sum = sum_type == "all_scans" or sum_type == "both"

        same_h5file = os.path.abspath(os.path.normpath(filename)) == os.path.abspath(
            os.path.normpath(output_filename)
        )
        if same_h5file:
            mode = "a"
        else:
            mode = "r"

        with h5py.File(filename, mode=mode) as h5infile:
            nscans = len(h5infile)
            scan_range = generate_range(start_scan, end_scan, nscans + 1)

            first_scan_name = list(h5infile.keys())[0]
            nimages, *detector_shape = h5infile[
                f"{first_scan_name}/measurement/{detector_name}"
            ].shape
            image_range = list(generate_range(start_image, end_image, nimages))

            scan_sum = ImageSum(detector_shape)
            full_sum = ImageSum(detector_shape)

            with h5py.File(output_filename, mode="a") as h5outfile:
                out_entry = h5outfile.require_group(output_entry)
                out_entry.attrs.setdefault("NX_class", "NXentry")
                out_entry.attrs["default"] = output_process
                out_process = out_entry.create_group(output_process)
                out_process.attrs["NX_class"] = "NXprocess"
                out_process.attrs["default"] = "results"
                out_results = out_process.create_group("results")
                out_results.attrs["NX_class"] = "NXdata"

                output_uris: list[str] = []
                monitor_uris: list[str] = []
                for scan_number in scan_range:
                    scan_images = h5infile[
                        f"{scan_number}.1/measurement/{detector_name}"
                    ]
                    scan_monitor = (
                        h5infile[f"{scan_number}.1/measurement/{monitor_name}"]
                        if monitor_name
                        else None
                    )
                    assert isinstance(scan_images, h5py.Dataset)
                    if scan_monitor is not None:
                        assert isinstance(scan_monitor, h5py.Dataset)

                    for image_number in image_range:
                        image = scan_images[image_number]
                        monitor = scan_monitor[image_number] if scan_monitor else None
                        scan_sum.add_to_sum(image, monitor)
                        full_sum.add_to_sum(image, monitor)

                        if save_scan_sums and scan_sum.nb_images == block_size:
                            name = f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}"
                            image_dset, monitor_dset = save_sum(
                                out_results, name=name, image_sum=scan_sum
                            )
                            output_uris.append(f"{output_filename}::{image_dset.name}")
                            if monitor_dset is not None:
                                monitor_uris.append(
                                    f"{output_filename}::{monitor_dset.name}"
                                )

                            # Move to next sum
                            scan_sum.reset()

                    if save_scan_sums and scan_sum.nb_images > 0:
                        name = f"Scan{scan_number}-Images{image_number - scan_sum.nb_images + 1}-{image_number}"
                        image_dset, monitor_dset = save_sum(
                            out_results, name=name, image_sum=scan_sum
                        )
                        output_uris.append(f"{output_filename}::{image_dset.name}")
                        if monitor_dset is not None:
                            monitor_uris.append(
                                f"{output_filename}::{monitor_dset.name}"
                            )

                if save_full_sum:
                    name = f"Sum of scans {scan_range.start} to {scan_range.stop - scan_range.step}"
                    image_dset, monitor_dset = save_sum(
                        out_results, name=name, image_sum=full_sum
                    )
                    output_uris.append(f"{output_filename}::{image_dset.name}")
                    if monitor_dset is not None:
                        monitor_uris.append(f"{output_filename}::{monitor_dset.name}")

        self.outputs.output_uris = output_uris
        if monitor_uris:
            self.outputs.monitor_uris = monitor_uris
        else:
            self.outputs.monitor_uris = None



import logging
from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import numpy

from .data_access import TaskWithDataAccess
from .utils import data_utils
from .utils.integrate_utils import is_counter_name

logger = logging.getLogger(__name__)


class BaseIntegrate(
    TaskWithDataAccess,
    input_names=[],
    optional_input_names=[
        "mask",
        "flatfield",
        "darkcurrent",
        "integration_options",
        "fixed_integration_options",
    ],
    register=False,
):

    def _get_ewoks_pyfai_options(self) -> dict:
        # Integration options
        ewoks_pyfai_options = dict()

        integration_options = data_utils.data_from_storage(
            self.inputs.integration_options, remove_numpy=True
        )
        if integration_options:
            ewoks_pyfai_options.update(integration_options)

        fixed_integration_options = data_utils.data_from_storage(
            self.get_input_value("fixed_integration_options", None), remove_numpy=True
        )
        if fixed_integration_options:
            ewoks_pyfai_options.update(fixed_integration_options)

        # Default units
        ewoks_pyfai_options.setdefault("unit", "2th_deg")

        # Correction images
        if not self.missing_inputs.mask and self.inputs.mask is not None:
            ewoks_pyfai_options["mask"] = self.get_image(
                data_utils.data_from_storage(self.inputs.mask)
            )
        if not self.missing_inputs.flatfield and self.inputs.flatfield is not None:
            ewoks_pyfai_options["flatfield"] = self.get_image(
                data_utils.data_from_storage(self.inputs.flatfield)
            )
        if not self.missing_inputs.darkcurrent and self.inputs.darkcurrent is not None:
            ewoks_pyfai_options["darkcurrent"] = self.get_image(
                data_utils.data_from_storage(self.inputs.darkcurrent)
            )

        return ewoks_pyfai_options

    def _pyfai_normalization_factor(
        self,
        monitors: List[Union[numpy.ndarray, Number, str, list, None]],
        references: List[Union[numpy.ndarray, Number, str, list, None]],
        ptdata: Optional[Dict[str, numpy.ndarray]] = None,
    ) -> Tuple[float, float]:
        r"""Returns the pyfai normalization factor based on a monitor and a references.

        The pyfai normalization factor is defined as

        .. code::

            Inorm = I / (normalization_factor1 * normalization_factor2 * ...)

        Monitor normalization is done like this

        .. code::

            Inorm = I * reference1 / monitor1 * reference2 / monitor2 * ...

        which means that the normalization factor is

        .. code::

            normalization_factor = monitor1 / reference1 * monitor2 / reference2 * ...

        Both monitors and references can be defined by:

         * scalar value
         * array or list of numbers
         * counter name
         * data URL
        """
        numerator = 1.0
        for value in monitors:
            if not data_utils.is_data(value):
                value = 1
            elif is_counter_name(value):
                if ptdata is None:
                    raise ValueError("monitor value cannot be a counter name")
                value = ptdata[value]
            else:
                value = self.get_data(value)
            numerator *= value

        if not numpy.isscalar(numerator):
            raise ValueError("monitor values need to be scalars")

        denominator = 1.0
        for value in references:
            if not data_utils.is_data(value):
                value = 1
            elif is_counter_name(value):
                if ptdata is None:
                    raise ValueError("reference value cannot be a counter name")
                value = ptdata[value]
            else:
                value = self.get_data(value)
            denominator *= value

        if not numpy.isscalar(denominator):
            raise ValueError("reference values need to be scalars")

        return numerator, denominator



from importlib.metadata import version
from packaging.version import Version

PYFAI_VERSION = Version(version("pyFAI"))

if PYFAI_VERSION >= Version("2025.1.0"):
    from pyFAI.integrator.azimuthal import AzimuthalIntegrator  # noqa F401
else:
    from pyFAI.azimuthalIntegrator import AzimuthalIntegrator  # noqa F401

if PYFAI_VERSION >= Version("2025.1.0"):
    from pyFAI.io.integration_config import (
        CURRENT_VERSION as INTEGRATION_SCHEMA_VERSION_MAX,
    )

    INTEGRATION_SCHEMA_VERSION_MIN = 1
elif PYFAI_VERSION >= Version("2024.9.0"):
    INTEGRATION_SCHEMA_VERSION_MAX = 4
    INTEGRATION_SCHEMA_VERSION_MIN = 1
else:
    INTEGRATION_SCHEMA_VERSION_MAX = 3
    INTEGRATION_SCHEMA_VERSION_MIN = 1



from __future__ import annotations

from typing import Sequence, Set, Union, List, Optional, Dict


def parse_fixed(
    fixed: Optional[List[str]], parametrization: Dict[str, List[str]]
) -> Set[str]:
    if not fixed:
        return set()
    allowed = {"dist", "poni1", "poni2", "rot1", "rot2", "rot3", "energy"}
    existing = set(fixed)
    unexpected = existing - allowed
    if unexpected:
        raise ValueError(f"'fixed' has unexpected parameters {sorted(unexpected)}")
    out = set()
    for param in fixed:
        out |= set(parametrization.get(param, {param}))
    return out


def parse_max_rings(max_rings: Union[int, Sequence[int], None]) -> Sequence[int | None]:
    if isinstance(max_rings, Sequence):
        return max_rings

    if not max_rings:
        return [None]

    if not isinstance(max_rings, int) or max_rings < 0:
        return [None]

    return [max_rings]



import json

import numpy
import pytest

from ..tasks.multi_geometry_integrate import Integrate2DMultiGeometry
from . import xrpd_theory


@pytest.fixture
def gonio_file(tmp_path):
    gonio_json = {
        "content": "Goniometer calibration v2",
        "detector": "Pilatus1M",
        "wavelength": 2.6379616687914948e-11,
        "param": [
            2.0062960356379422,
            0.015883422992655393,
            0.8216006393975213,
            0.002164173240804926,
            -0.0031180017089310853,
            0.0063336570464145835,
            0.004297460651999008,
        ],
        "param_names": ["dist", "poni1", "poni2", "rot1", "rot2", "rot_x", "rot_y"],
        "pos_names": ["pos"],
        "trans_function": {
            "content": "GeometryTransformation",
            "param_names": ["dist", "poni1", "poni2", "rot1", "rot2", "rot_x", "rot_y"],
            "pos_names": ["pos"],
            "dist_expr": "dist",
            "poni1_expr": "poni1 + rot_x*cos(pos) - rot_y*sin(pos)",
            "poni2_expr": "poni2 + rot_x*sin(pos) + rot_y*cos(pos)",
            "rot1_expr": "+rot1*cos(pos) - rot2*sin(pos)",
            "rot2_expr": "+rot1*sin(pos) + rot2*cos(pos)",
            "rot3_expr": "pos",
            "constants": {"pi": 3.141592653589793},
        },
    }
    filename = tmp_path / "gonio.json"
    with open(filename, "w") as f:
        json.dump(gonio_json, f)

    return filename


def test_multi_geometry_task(imageSetup1Calibrant1, gonio_file):
    assert isinstance(imageSetup1Calibrant1, xrpd_theory.Calibration)
    images = numpy.zeros(shape=(10, *imageSetup1Calibrant1.image.shape))
    for i in range(10):
        images[i] = imageSetup1Calibrant1.image

    task = Integrate2DMultiGeometry(
        inputs={
            "goniometer_file": gonio_file,
            "integration_options": {"npt_rad": 300, "npt_azim": 180},
            "images": images,
            "positions": [*numpy.linspace(0, 360, 10)],
        }
    )

    task.execute()

    assert task.get_output_value("radial").shape == (300,)
    assert task.get_output_value("azimuthal").shape == (180,)
    assert task.get_output_value("intensity").shape == (180, 300)



import logging
from contextlib import contextmanager
from collections import OrderedDict
from typing import Iterable, Dict, Mapping
from packaging.version import Version

import pyFAI
import pyFAI.worker
from ewokscore.hashing import uhash

from .. import pyfai_api
from .utils import pyfai_utils

_WORKER_POOL = None


logger = logging.getLogger(__name__)


class WorkerPool:
    """Pool with one worker per configuration up to a maximum number of workers."""

    def __init__(self, nworkers: int = 1) -> None:
        self._workers: Dict[int, pyFAI.worker.Worker] = OrderedDict()
        self.nworkers = nworkers

    @staticmethod
    def _worker_id(*args):
        return uhash(args)

    @property
    def nworkers(self):
        return self._nworkers

    @nworkers.setter
    def nworkers(self, value: int):
        self._nworkers = value
        self._check_pool_size()

    def _check_pool_size(self):
        while self._workers and len(self._workers) > self.nworkers:
            self._workers.popitem(last=False)

    @contextmanager
    def worker(
        self, ewoks_pyfai_options: Mapping, demo: bool = False
    ) -> Iterable[pyFAI.worker.Worker]:
        # TODO: deal with threads and subprocesses
        worker_options, integration_options = (
            pyfai_utils.split_worker_and_integration_options(ewoks_pyfai_options)
        )
        logger.info("Pyfai worker options: %s", worker_options)
        logger.info("Pyfai integration options: %s", integration_options)

        worker_id = self._worker_id(worker_options, integration_options, demo)
        worker = self._workers.pop(worker_id, None)
        if worker is None:
            logger.info("Creating a new pyfai worker")
            worker = self._create_worker(worker_options, integration_options, demo)
        self._workers[worker_id] = worker

        self._check_pool_size()
        logger.info("Pyfai integration method: %s", worker._method)
        yield worker

    @staticmethod
    def _create_worker(
        worker_options: Mapping, integration_options: Mapping, demo: bool
    ) -> pyFAI.worker.Worker:
        if demo:
            return DemoWorker(integration_options, worker_options)
        return EwoksWorker(integration_options, worker_options)


def _get_global_pool() -> WorkerPool:
    global _WORKER_POOL
    if _WORKER_POOL is None:
        _WORKER_POOL = WorkerPool()
    return _WORKER_POOL


def set_maximum_persistent_workers(nworkers: int) -> None:
    pool = _get_global_pool()
    pool.nworkers = nworkers


class EwoksWorker(pyFAI.worker.Worker):
    def __init__(self, integration_options: Mapping, worker_options: Mapping) -> None:
        super().__init__(**worker_options)
        self.output = "raw"
        self._i = 0

        integration_options, mask, flatfield, darkcurrent = (
            pyfai_utils.extract_images_from_integration_options(integration_options)
        )

        provided = set(integration_options)
        self.set_config(integration_options, consume_keys=True)
        unused = {k: v for k, v in integration_options.items() if k in provided}
        if unused:
            logger.warning("Unused pyFAI integration options: %s", unused)
        else:
            logger.info("All pyFAI integration options were used")

        # Flat/dark correction:
        #   Icor = (I - darkcurrent) / flatfield
        if mask is not None:
            self.ai.detector.set_mask(mask)
        if flatfield is not None:
            self.ai.detector.set_flatfield(flatfield)
        if darkcurrent is not None:
            self.ai.detector.set_darkcurrent(darkcurrent)

    def set_config(self, integration_options, consume_keys=False):
        if pyfai_api.PYFAI_VERSION < Version("2025.1.0"):
            # In older pyFAI versions, the "method" is required
            # when upgrading version 2 to 3.
            version = integration_options.get("version", 1)

            if version == 2:
                if self.integrator_name and "clip" in self.integrator_name:
                    method = "csr"
                else:
                    method = ""
                integration_options["method"] = method

        super().set_config(integration_options, consume_keys=consume_keys)


class DemoWorker(EwoksWorker):
    def process(self, data, *args, **kwargs):
        return super().process(data[:-1, :-1], *args, **kwargs)


@contextmanager
def persistent_worker(
    ewoks_pyfai_options: Mapping, demo: bool = False
) -> Iterable[pyFAI.worker.Worker]:
    """Get a worker for a particular configuration that stays in memory."""
    pool = _get_global_pool()
    with pool.worker(ewoks_pyfai_options, demo) as worker:
        yield worker



import time
import logging
from contextlib import contextmanager
from typing import Iterator, Optional, Any, Tuple, Generator

from ewoksdata.data.contextiterator import contextiterator

_logger = logging.getLogger(__name__)


@contextiterator
def zip_with_progress(
    *iterators: Iterator[Iterator[Any]],
    message: str = "Progress %d/%s",
    nmax: Optional[int] = None,
    period: float = 5,
    logger=None,
    iter_time_name: str = "iteration",
) -> Generator[Tuple[Any], None, None]:
    """Like python's zip but will progress logging when iterating over the result.
    In addition it yields a context manager which allows for timing different sections
    in the context as well as the yielding time itself:

    .. code::python

        for v1, v2, tmcontext in zip_with_progress([...], [...], iter_time_name="read"):
            with tmcontext("process"):
                ...
            with tmcontext("write"):
                ...
    """
    ctimer = _ContextTimer(iter_time_name)
    if logger is None:
        logger = _logger
    i = 0
    t0 = time.time()
    if nmax is None:
        nmax = "?"
    try:
        for tpl in zip(*iterators):
            yield (*tpl, ctimer.tmcontext)
            i += 1
            if (time.time() - t0) > period:
                t0 = time.time()
                _logger.info(f"{message} (ONGOING: {ctimer})", i, nmax)
    finally:
        _logger.info(f"{message} (FINISHED: {ctimer})", i, nmax)


class _ContextTimer:
    def __init__(self, rest_context: str = "other") -> None:
        self._total_t0 = time.perf_counter_ns()
        self._context_time = dict()
        self._rest_context = rest_context

    @contextmanager
    def tmcontext(self, name: str) -> Generator[None, None, None]:
        context_t0 = time.perf_counter_ns()
        try:
            yield
        finally:
            context_t1 = time.perf_counter_ns()
            self._context_time.setdefault(name, 0)
            self._context_time[name] += context_t1 - context_t0

    def __str__(self) -> str:
        dt_total = time.perf_counter_ns() - self._total_t0
        dt_sum = sum([dt for dt in self._context_time.values()])
        dt_rest = dt_total - dt_sum
        dts = [
            f"{name}={self.strftime(dt, dt_total)}"
            for name, dt in self._context_time.items()
        ]
        dts.append(f"{self._rest_context}={self.strftime(dt_rest, dt_total)}")
        dts.append(f"TOTAL={self.strftime(dt_total, dt_total)}")
        return ", ".join(dts)

    @staticmethod
    def strftime(dt_ns: int, dt_total: int) -> str:
        m, s = divmod(int(dt_ns * 1e-9 + 0.5), 60)
        h, m = divmod(m, 60)
        f = int(dt_ns / dt_total * 100 + 0.5)
        return f"{h:02d}:{m:02d}:{s:02d} ({f:d}%%)"



from pyFAI.goniometer import Goniometer
from typing import Tuple
from pyFAI.units import Unit
from .base_integrate import BaseIntegrate
from .utils import integrate_utils, pyfai_utils


class Integrate2DMultiGeometry(
    BaseIntegrate,
    input_names=["goniometer_file", "positions", "images"],
    output_names=[
        "info",
        "radial",
        "azimuthal",
        "intensity",
        "intensity_error",
        "radial_units",
        "azimuthal_units",
    ],
):
    def run(self):
        raw_integration_options = self._get_ewoks_pyfai_options()
        multi_geometry_options, integrate2d_options = (
            pyfai_utils.split_multi_geom_and_integration_options(
                raw_integration_options
            )
        )

        goniometer = Goniometer.sload(self.inputs.goniometer_file)
        mg_ai = goniometer.get_mg(self.inputs.positions, **multi_geometry_options)

        result = mg_ai.integrate2d(self.inputs.images, **integrate2d_options)

        self.outputs.radial = result.radial
        self.outputs.azimuthal = result.azimuthal
        self.outputs.intensity = result.intensity
        self.outputs.intensity_error = integrate_utils.get_intensity_error(result)

        result_unit: Tuple[Unit, Unit] = result.unit
        radial_unit, azim_unit = result_unit
        self.outputs.radial_units = radial_unit.name
        self.outputs.azimuthal_units = azim_unit.name

        self.outputs.info = pyfai_utils.compile_integration_info(
            raw_integration_options
        )



from contextlib import contextmanager
from typing import Iterator, Union
from packaging.version import Version
from importlib.metadata import version

import numpy
import h5py
from silx.io.url import DataUrl
from silx.io import h5py_utils
from ewokscore import TaskWithProgress
from ewoksdata.data import nexus
from ewoksdata.data import bliss
from ewoksdata.data.contextiterator import contextiterator

from .utils import data_utils


_LIMA_TEMPLATE_SUPPORTED = Version(version("blissdata")) >= Version("1.1.0")


class TaskWithDataAccess(
    TaskWithProgress,
    optional_input_names=[
        "retry_timeout",
        "retry_period",
        "lima_url_template",
        "lima_url_template_args",
    ],
    register=False,
):
    def get_retry_options(self):
        retry_timeout = self.get_input_value("retry_timeout", None)
        retry_period = self.get_input_value("retry_period", None)
        return {"retry_timeout": retry_timeout, "retry_period": retry_period}

    def _get_blissdata_options(self):
        lima_url_template = self.get_input_value("lima_url_template", None)
        lima_url_template_args = self.get_input_value("lima_url_template_args", None)
        if _LIMA_TEMPLATE_SUPPORTED:
            return {
                "lima_url_template": lima_url_template,
                "lima_url_template_args": lima_url_template_args,
                **self.get_retry_options(),
            }
        if lima_url_template:
            raise ValueError("'lima_url_template' requires blissdata>=1.1.0")
        if lima_url_template_args:
            raise ValueError("'lima_url_template_args' requires blissdata>=1.1.0")
        return self.get_retry_options()

    @contextmanager
    def open_h5item(
        self,
        url: Union[str, DataUrl],
        create: bool = False,
        overwrite: bool = False,
        **openoptions,
    ) -> Iterator[Union[h5py.Group, h5py.Dataset, numpy.ndarray]]:
        if not isinstance(url, DataUrl):
            url = DataUrl(url)
        retryoptions = self.get_retry_options()
        if create:
            url = nexus.create_url(url, overwrite=overwrite, **retryoptions)
        with h5py_utils.open_item(
            url.file_path(), url.data_path(), **retryoptions, **openoptions
        ) as item:
            idx = url.data_slice()
            if idx is None:
                yield item
            else:
                yield item[idx]

    def get_data(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_data(*args, **kw)

    def get_image(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_image(*args, **kw)

    @contextiterator
    def iter_bliss_data(self, *args, **kw):
        kw.update(self._get_blissdata_options())
        with bliss.iter_bliss_scan_data(*args, **kw) as iterator:
            yield from iterator

    def iter_bliss_data_from_memory(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data_from_memory(*args, **kw)

    def iter_bliss_scan_data_from_memory_slice(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data_from_memory_slice(*args, **kw)

    def link_bliss_scan(self, *args, **kw):
        kw.update(self._get_blissdata_options())
        return data_utils.link_bliss_scan(*args, **kw)



from __future__ import annotations
import logging
import warnings
from numbers import Number
from typing import Dict, List, Optional, Set, Tuple, Sequence, Union

import pyFAI.detectors
import pyFAI.calibrant
import pyFAI.geometry
import pyFAI.goniometer
import pyFAI.geometryRefinement

from .data_access import TaskWithDataAccess
from .utils import calibrate_utils
from .utils import data_utils
from .utils import xrpd_utils


__all__ = ["CalibrateSingle", "CalibrateMulti", "CalculateGeometry"]


logger = logging.getLogger(__name__)


class CalibrateSingle(
    TaskWithDataAccess,
    input_names=["image", "detector", "calibrant", "geometry", "energy"],
    optional_input_names=[
        "detector_config",
        "fixed",
        "max_rings",
        "robust",
        "ring_detector",
    ],
    output_names=["geometry", "energy", "detector", "detector_config", "rings", "chi2"],
):
    """Single distance and energy calibration. Accepts multiple values for max_rings for sequential refinement."""

    def run(self):
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)
        wavelength0 = xrpd_utils.energy_wavelength(self.inputs.energy)

        # Find diffraction rings and fit
        geometry_refinement = self._refine(
            dict(geometry0),
            wavelength0,
            calibrant_object,
            detector_object,
        )

        # Parse results
        results = geometry_refinement.get_config()
        geometry = {
            k: v for k, v in results.items() if k in xrpd_utils.GEOMETRY_PARAMETERS
        }
        energy = xrpd_utils.energy_wavelength(results["wavelength"])
        rings = xrpd_utils.points_to_rings(geometry_refinement.data, calibrant_object)

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        if geometry_refinement.data.ndim != 2:
            self.outputs.chi2 = float("nan")
        else:
            self.outputs.chi2 = geometry_refinement.chi2()

    def _refine(
        self,
        geometry0: dict,
        wavelength0: float,
        calibrant: pyFAI.calibrant.Calibrant,
        detector: pyFAI.detectors.Detector,
    ) -> pyFAI.geometryRefinement.GeometryRefinement:
        """Find diffraction rings and fit"""
        parametrization = {"energy": ["wavelength"]}
        always_fixed_params = calibrate_utils.parse_fixed(
            self.inputs.fixed, parametrization
        )
        all_parameters = xrpd_utils.GEOMETRY_PARAMETERS | {"wavelength"}

        if self.inputs.robust:
            incremental_release_groups = [
                {"dist"},
                {"poni1", "poni2"},
                {"rot1", "rot2", "rot3"},
                {"wavelength"},
            ]
        else:
            incremental_release_groups = [set(all_parameters)]

        # Initial fit parameters
        geometry = geometry0
        wavelength = wavelength0

        max_rings_list = calibrate_utils.parse_max_rings(self.inputs.max_rings)
        fixed_parameter_sets = _fixed_parameters_sets(
            all_parameters, always_fixed_params, incremental_release_groups
        )
        n_rings = len(max_rings_list)
        n_releases = len(fixed_parameter_sets)
        for i, max_rings in enumerate(max_rings_list):
            for j, fixed_parameters in enumerate(fixed_parameter_sets):
                # Detect experimental rings
                cps = self._extract_control_points(
                    geometry, calibrant, wavelength, detector, max_rings
                )
                # Fit parameters to detected rings
                setup = pyFAI.goniometer.SingleGeometry(
                    "single_user_detector",
                    detector=detector,
                    geometry=geometry,
                    calibrant=calibrant,
                    control_points=cps,
                )
                setup.set_wavelength(wavelength)

                if cps is None:
                    logger.warning("No rings detected for calibration")
                    return setup.geometry_refinement

                _refine_geometry(
                    setup.geometry_refinement, fixed_parameters=fixed_parameters
                )

                # Update parameters with fit result
                results = setup.geometry_refinement.get_config()
                geometry = {
                    k: v
                    for k, v in results.items()
                    if k in xrpd_utils.GEOMETRY_PARAMETERS
                }
                wavelength = results["wavelength"]

                self.progress = (i * n_rings + j) / (n_releases * n_rings) * 100
        return setup.geometry_refinement

    def _extract_control_points(
        self,
        geometry: dict,
        calibrant: pyFAI.calibrant.Calibrant,
        wavelength: float,
        detector: pyFAI.detectors.Detector,
        max_rings: int | None,
    ) -> pyFAI.goniometer.ControlPoints | None:
        ring_detector_name = self.get_input_value("ring_detector", None)
        if ring_detector_name is None:
            ring_detector = detector
        else:
            ring_detector = pyFAI.detectors.detector_factory(ring_detector_name)

        setup = pyFAI.goniometer.SingleGeometry(
            "single_ring_detector",
            image=self.get_image(self.inputs.image),
            detector=ring_detector,
            geometry=geometry,
            calibrant=calibrant,
        )
        setup.set_wavelength(wavelength)

        setup.extract_cp(max_rings=max_rings)
        if setup.geometry_refinement.data.ndim != 2:
            logger.warning("No rings detected for calibration")
            return

        return setup.control_points


class CalibrateMulti(
    TaskWithDataAccess,
    input_names=[
        "images",
        "positions",
        "detector",
        "calibrant",
        "geometry",
        "energy",
    ],
    optional_input_names=[
        "reference_position",
        "sample_position",
        "positionunits_in_meter",
        "detector_config",
        "fixed",
        "max_rings",
        "robust",
    ],
    output_names=[
        "geometry",
        "energy",
        "detector",
        "detector_config",
        "rings",
        "chi2",
        "parametrization",
        "parameters",
    ],
):
    """Single energy, multi distance calibration with 1 distance motor.

    The `images` and `positions` are the diffraction patterns and the motor positions at which those were measured.

    The `reference_position` is the detector position that corresponds to `geometry` (`positions[0]` by default).

    The `sample_position` is the position where sample and detector would theoretically coincide (0 by default).

    The output `energy` and `geometry` correspond to `sample_position`.

    The units of `positions`, `sample_position` and `reference_position` are defined by `positionunits_in_meter`
    (1e-3 by default which means the unit is in millimeter).

    The parametrization allow deriving energy+geometry from the detector position.

    .. code-block:: python

        energy, geometry = parametrization(parameters, position)
    """

    def run(self):
        if len(self.inputs.images) != len(self.inputs.positions):
            raise ValueError("number of 'images' and 'positions' must be the same")
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        energy0 = self.inputs.energy
        wavelength0 = xrpd_utils.energy_wavelength(energy0)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)

        # Parameterize the detector geometry
        trans_function, initial_guess = self._reparametrization(geometry0, energy0)

        # Define setup
        setup = pyFAI.goniometer.GoniometerRefinement(
            initial_guess,
            pos_function=lambda x: x,
            trans_function=trans_function,
            detector=detector_object,
            wavelength=wavelength0,
        )
        for i, (image, xi) in enumerate(zip(self.inputs.images, self.inputs.positions)):
            label = f"position{i}"
            image = self.get_image(image)
            xi = self.get_data(xi)
            setup.new_geometry(
                label, image=image, metadata=xi, calibrant=calibrant_object
            )

        # Find diffraction rings and fit
        self._refine(setup)

        # Parse results
        geometry, energy = self._get_fit_results(setup)
        paramerization = self._get_parameterization(setup)
        parameters = self._get_parameters(setup)
        rings = self._get_rings(setup, calibrant_object)
        chi2 = setup.chi2()

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        self.outputs.chi2 = chi2
        self.outputs.parametrization = paramerization
        self.outputs.parameters = parameters

    def _reparametrization(
        self, geometry0: Dict, energy0: float
    ) -> Tuple[pyFAI.goniometer.ExtendedTransformation, List[Number]]:
        # xunits are the position units
        # dunits are the distance units of the fit parameters
        # aunits are the angular units of the fit parameters
        dunits_per_m = 100  # dunits/m
        aunits_per_rad = 10  # aunits/rad
        m_per_xunits = self._get_position_units()  # m/xunits

        trans_function = pyFAI.goniometer.ExtendedTransformation(
            param_names=[
                "dist_offset",  # dunits
                "dist_scale",  # dunits/xunits
                "poni1_offset",  # dunits
                "poni1_scale",  # dunits/xunits
                "poni2_offset",  # dunits
                "poni2_scale",  # dunits/xunits
                "arot1",  # aunits
                "arot2",  # aunits
                "arot3",  # aunits
                "energy",  # keV
            ],
            pos_names=["x"],  # xunits
            dist_expr=f"dist_scale / {dunits_per_m} * x + dist_offset / {dunits_per_m}",
            poni1_expr=f"poni1_scale / {dunits_per_m} * x + poni1_offset / {dunits_per_m}",
            poni2_expr=f"poni2_scale / {dunits_per_m} * x + poni2_offset / {dunits_per_m}",
            rot1_expr=f"arot1 / {aunits_per_rad}",
            rot2_expr=f"arot2 / {aunits_per_rad}",
            rot3_expr=f"arot3 / {aunits_per_rad}",
            wavelength_expr="hc*1e-10/energy",
        )

        xref = self._get_reference_position()  # xunits
        dist_offset = self._get_sample_position() * m_per_xunits  # m
        dist_scale = (geometry0["dist"] - dist_offset) / xref  # m/xunits

        poni1_scale = 0.0  # m/xunits
        poni1_offset = geometry0["poni1"]  # m
        poni2_scale = 0.0  # m/xunits
        poni2_offset = geometry0["poni2"]  # m

        initial_guess = [
            dist_offset * dunits_per_m,
            dist_scale * dunits_per_m,
            poni1_offset * dunits_per_m,
            poni1_scale * dunits_per_m,
            poni2_offset * dunits_per_m,
            poni2_scale * dunits_per_m,
            geometry0["rot1"] * aunits_per_rad,
            geometry0["rot2"] * aunits_per_rad,
            geometry0["rot3"] * aunits_per_rad,
            energy0,
        ]
        return trans_function, initial_guess

    @property
    def _max_rings(self):
        max_rings = self.inputs.max_rings
        if not max_rings:
            return None

        if max_rings < 0:
            return None

        return max_rings

    def _refine(self, setup: pyFAI.goniometer.GoniometerRefinement) -> None:
        """Find diffraction rings and fit"""
        parametrization = {
            "dist": ["dist_offset", "dist_scale"],
            "poni1": ["poni1_offset", "poni1_scale"],
            "poni2": ["poni2_offset", "poni2_scale"],
            "rot1": ["arot1"],
            "rot2": ["arot2"],
            "rot3": ["arot3"],
        }
        always_fixed_params = calibrate_utils.parse_fixed(
            self.inputs.fixed, parametrization
        )
        all_parameters = set(setup.trans_function.param_names)

        if self.inputs.robust:
            incremental_release_groups = [
                {"dist_offset", "dist_scale"},
                {"poni1_offset", "poni2_offset"},
                {"arot1", "arot2", "arot3"},
                {"poni1_scale", "poni2_scale"},
                {"energy"},
            ]
        else:
            incremental_release_groups = [set(all_parameters)]

        fixed_parameter_sets = _fixed_parameters_sets(
            all_parameters, always_fixed_params, incremental_release_groups
        )
        n_releases = len(fixed_parameter_sets)
        for i, fixed_parameters in enumerate(fixed_parameter_sets, 1):
            # Detect experimental rings
            for setup_single_dist in setup.single_geometries.values():
                setup_single_dist.extract_cp(max_rings=self._max_rings)
                if setup_single_dist.geometry_refinement.data.ndim != 2:
                    logger.warning("No rings detected for calibration")
                    return

            # Fit parameters to detected rings
            _refine_geometry(setup, fixed_parameters=fixed_parameters)

            self.progress = i / n_releases * 100

    def _get_rings(
        self,
        setup: pyFAI.goniometer.GoniometerRefinement,
        calibrant: pyFAI.calibrant.Calibrant,
    ) -> Dict[str, dict]:
        rings = dict()
        for i, setup_single_dist in enumerate(setup.single_geometries.values()):
            points = setup_single_dist.geometry_refinement.data
            rings[str(i)] = xrpd_utils.points_to_rings(points, calibrant)
        return rings

    def _get_fit_results(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> Tuple[dict, Number]:
        parameters = setup.param
        xref = self._get_reference_position()
        results = setup.trans_function(parameters, xref)
        geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
        energy = xrpd_utils.energy_wavelength(results.wavelength)
        return geometry, energy

    def _get_parameterization(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> dict:
        parametrization = dict(setup.trans_function.to_dict())
        parametrization.pop("constants")
        parametrization.pop("content")
        return parametrization

    def _get_parameters(self, setup: pyFAI.goniometer.GoniometerRefinement) -> dict:
        parameters = setup.param
        names = setup.trans_function.param_names
        return dict(zip(names, parameters))

    def _get_reference_position(self):
        """first image position by default (xunits)"""
        if self.missing_inputs.reference_position:
            return self.get_data(self.inputs.positions[0])
        else:
            return self.get_data(self.inputs.reference_position)

    def _get_sample_position(self):
        """0 by default (xunits)"""
        if self.missing_inputs.sample_position:
            return 0
        else:
            return self.get_data(self.inputs.sample_position)

    def _get_position_units(self):
        """millimeter by default"""
        if self.inputs.positionunits_in_meter:
            return self.inputs.positionunits_in_meter  # m/xunits
        else:
            return 1e-3  # m/xunits


def calculate_geometry(
    parametrization: dict, parameters: dict, position: Number
) -> Tuple[dict, Number]:
    trans_function = pyFAI.goniometer.ExtendedTransformation(**parametrization)
    parameters = [parameters[name] for name in trans_function.param_names]
    trans_function(parameters, position)
    results = trans_function(parameters, position)
    geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
    energy = xrpd_utils.energy_wavelength(results.wavelength)
    return geometry, energy


class CalculateGeometry(
    TaskWithDataAccess,
    input_names=["parametrization", "parameters", "position"],
    output_names=["geometry", "energy"],
):
    """Calculate energy and geometry from pyFAI parametrization"""

    def run(self):
        parametrization = data_utils.data_from_storage(self.inputs.parametrization)
        geometry, energy = calculate_geometry(
            parametrization,
            self.inputs.parameters,
            self.get_data(self.inputs.position),
        )
        self.outputs.geometry = geometry
        self.outputs.energy = energy


def _fixed_parameters_sets(
    all_params: Set[str],
    always_fixed_params: Set[str],
    incremental_release_groups: Sequence[Set[str]],
) -> List[Optional[Set[str]]]:
    """
    Returns a sequence of fixed parameter sets based on incremental parameter releases.

    :param all_params: The complete set of parameters.
    :param always_fixed_params: Parameters that are always fixed, regardless of release groups.
    :param incremental_release_groups: A sequence of parameter groups that are released incrementally.
    :returns: A list where each element is a set of parameters that remain fixed
              after the corresponding release step.
    """
    fixed_parameters = list()
    accumulate_released_parameters = set()
    for release_group in incremental_release_groups:
        accumulate_released_parameters |= release_group - always_fixed_params
        if not accumulate_released_parameters:
            continue
        fixed = all_params - accumulate_released_parameters
        if not fixed:
            fixed = None
        if fixed in fixed_parameters:
            continue
        fixed_parameters.append(fixed)
    return fixed_parameters


def _refine_geometry(
    geometry_refinement: Union[
        pyFAI.goniometer.GoniometerRefinement,
        pyFAI.geometryRefinement.GeometryRefinement,
    ],
    fixed_parameters: Optional[Sequence] = None,
) -> bool:
    """
    Refines the given geometry within the default parameter bounds.

    :param geometry_refinement: A pyFAI object that parameterizes a single or multi-distances geometry.
    :param fixed_parameters: Parameters that are not allowed to change during refinement.
    :return: True if refinement is successful within the default parameter bounds, False otherwise.
    """
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.filterwarnings(
            "always",
            message="Values in x were outside bounds during a minimize step, clipping to bounds",
            category=RuntimeWarning,
        )
        geometry_refinement.refine3(fix=fixed_parameters)  # method="slsqp"
        if not caught_warnings:
            return True

    return False
